import torch
from torch import nn
import torch.nn.functional as F


class SepBN(nn.Module):

    def __init__(self, inplanes, pool_size=3, num_groups=2, sep_ways=3, expansion=2, temp=30):
        super(SepBN, self).__init__()
        self.num_groups = num_groups
        self.sep_ways = sep_ways
        self.temp = temp

        self.bn = nn.BatchNorm2d(inplanes, affine=False)

        self.weight_ = nn.Parameter(torch.ones(sep_ways,1,inplanes,1,1))
        self.bias_ = nn.Parameter(torch.zeros(sep_ways,1,inplanes,1,1))

        self.avgpool = nn.AdaptiveAvgPool2d(pool_size)
        self.conv = nn.Conv2d(inplanes, num_groups, 1)
        self.fc1 = nn.Linear(int(num_groups * pool_size * pool_size), int(num_groups * sep_ways * expansion))
        self.fc2 = nn.Linear(int(num_groups * sep_ways * expansion),  int(num_groups * sep_ways))

        for m in self.modules():
            if isinstance(m, (nn.Linear, nn.Conv2d)):
                m.reset_parameters()
        nn.init.normal_(self.weight_, 1., 0.1)
        nn.init.normal_(self.bias_, 0., 0.1)

    def updata_temperature(self):
        if self.temp != 1:
            self.temp -= 1

    def forward(self, x):
        batchsize, channel, height, width = x.size()

        out = self.avgpool(x)
        out = self.conv(out)
        out = F.leaky_relu(out)
        out = self.fc1(out.reshape(batchsize, -1))
        out = F.leaky_relu(out)
        out = self.fc2(out).view(batchsize, self.num_groups, -1)
        out = F.softmax(out/self.temp, 2)
        out = out.permute(2,0,1).contiguous().unsqueeze(-1).unsqueeze(-1).unsqueeze(-1)

        x = self.bn(x)
        x = x.unsqueeze(0)

        x = x * self.weight_ + self.bias_
        x = x.reshape(self.sep_ways, batchsize, self.num_groups, -1, height, width)

        out = torch.sum(x * out, dim=0).reshape(batchsize, channel, height,width)
        return out